//*****************************************************************************
// Description     : histogram measurement program
// Target Device   : APG7400A
// Target Platform : Linux, C
// Date            : October 2017
// Version         : 1.0.1  2024/1/9  ADG MAX 16384->4096 
//*****************************************************************************
#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>
#include    <sys/stat.h>
#include    <sys/types.h>
#include    <sys/socket.h>
#include    <netinet/in.h>

#include    "ftd2xx.h"
#include    "usbmca.h"

#include    "mdebug.h"

/******************************************************************************/
#define     MEASURING_TIME      (3000llu)   /* msec */
#define     SAVE_FOLDER         "/tmp/usbmca/"
#define     SAVE_NAME           "hist.csv"
#define     SAVE_FILE           (SAVE_FOLDER SAVE_NAME)
/******************************************************************************/
static int setting_config(FT_HANDLE usb_handle, int *adg) ;
static int run_measure(FT_HANDLE usb_handle, USBMCA_HISTOGRAM *histdt, int *adg) ;


/* #########################################################################
    Name      : main
    Function  : The main entry point for the application.
    Arguments :
        in argc : number of arguments
        in argv : value  of argument
    Return :
         1 : success
   ######################################################################### */
int main(int argc, char **argv)
{
    USBMCA_HISTOGRAM    histdt;
    FT_STATUS   ret0th, ret1st, ret2nd ;
    FT_HANDLE   usb_handle ;
    DWORD   oldid[2] ;
    int     i, adg[USBMCA_MAX_CHNUM] ;
    int     count ;
    int     result ;
    int     erron, errcd ;
    int     old_vid, old_pid;
    char    *serialnum ;

    mkdir(SAVE_FOLDER, 0777);   /* create directory of save file */
    unlink(SAVE_FILE);          /* delete save file */

    erron = 0 ;
    usb_handle = NULL ;
    
    serialnum = USBMCA_SERIALNUM ;

    memset(adg, 0, sizeof(adg));
    memset(&histdt, 0, sizeof(histdt));

    /********** setting id(vendor, product) **********/
    old_vid = 0 ;
    old_pid = 0 ;
    ret1st = usbmca_get_vidpid(&old_vid, &old_pid) ;
    ret2nd = usbmca_set_vidpid(USBMCA_ID_VENDOR, USBMCA_ID_PRODUCT) ;
    
    if((ret1st | ret2nd) != FT_OK) {
        printf("setting ID(vendor + product) error(%lld)...\n", ret1st) ;
        erron = -11 ;
    }
    
    /*** device connect(by serial number) ***/
    if(erron == 0) {
		ret0th = usbmca_open_by_serialnum(serialnum, &usb_handle);
        
        if(ret0th != FT_OK) {
            printf("handle open error(%s, %lld)...\n", serialnum, ret0th) ;
            erron = -31 ;
        }
    }
    
    /*** device initial ***/
    if(erron == 0) {
        ret0th = usbmca_initial(usb_handle) ;
        
        if(ret0th != FT_OK) {
            printf("device initial error(%s, %lld)...\n", serialnum, ret0th) ;
            erron = -41 ;
        }

        msleep(500);   /* wait initial complete.. */
    }
    
    /*** setting config ***/
    if(erron == 0) {
        result = setting_config(usb_handle, adg) ;
        
        if(result < 0) {
            printf("device setting config error(%s, %d)...\n", serialnum, result) ;
            erron = -51 ;
        }
    }
    
    /*** run measurement ***/
    if(erron == 0) {
        result = run_measure(usb_handle, &histdt, adg);

        if (result == 0) {
            printf("complete!!!\n");
        }
        else {
            printf("measuring error(%s, %d)...\n", serialnum, result);
            erron = -61;
        }
    }
    
    /*** device disconnect ***/
    if(usb_handle != 0) {
        usbmca_close(usb_handle) ;
    }
    
    /********** restore id(vendor, product) **********/
    usbmca_set_vidpid(old_vid, old_pid) ;
    
    return  (EXIT_SUCCESS) ;
    
}


/* #########################################################################
    Name      : setting_config
    Function  : setting the measurement configuration
    Arguments :
        in usb_handle : USB handle
        in adg        : ADC gain
    Return :
          0 : success
        ~-1 : write error
   ######################################################################### */
static int setting_config(FT_HANDLE usb_handle, int *adg)
{
    FT_STATUS   ret[128] ;
    int     result ;
    int     i, index ;
    static const int chn_list[] = {
        USBMCA_CMDTYPE_CH1, USBMCA_CMDTYPE_CH2,
        USBMCA_CMDTYPE_CH3, USBMCA_CMDTYPE_CH4
    };
    
    result = 0 ;
    
    for(i = 0 ; i < (int)(_countof(ret)) ; i++) {
        ret[i] = FT_OK ;
    }
    
    /* adc gain setting */
    for (i = 0; i < USBMCA_MAX_CHNUM; i++) {
        adg[i] = USBMCA_ADG_04K;    /* MAX 4096ch */
    }
    
    index = 0 ;

    /*****config, all CH*****/
    for (i = 0; i < (int)(_countof(chn_list)); i++) {
        /* ADC gain = 8192ch */
        ret[index++] = usbmca_order_command_lite(usb_handle, chn_list[i], USBMCA_CMD_ADG, (unsigned int)adg[i]);
        /* threshold = 80 ch */
        ret[index++] = usbmca_order_command_lite(usb_handle, chn_list[i], USBMCA_CMD_THR, 80);
        /* lower level discrimination = 95ch */
        ret[index++] = usbmca_order_command_lite(usb_handle, chn_list[i], USBMCA_CMD_LLD, 95);
        /* upper level discrimination = 8191ch */
        ret[index++] = usbmca_order_command_lite(usb_handle, chn_list[i], USBMCA_CMD_ULD, 8191);
        /* offset level = 0ch */
        ret[index++] = usbmca_order_command_lite(usb_handle, chn_list[i], USBMCA_CMD_OFS, 0);
    }
    
    /* PDS: absolution time */
    ret[index++] = usbmca_order_command_lite(usb_handle, USBMCA_CMDTYPE_CMN, USBMCA_CMD_PDSW, USBMCA_PDS_ABS);

    for(i = 0 ; i < index ; i++) {
        if(ret[i] != FT_OK) {
            result = (i + 1) * (-1) ;
            break ;
        }
    }
    
    return  result ;
    
}


/* #########################################################################
    Name      : run_measure
    Function  : measurement running
    Arguments :
        in  usb_handle : USB handle
        out histdt     : histogram data
        in  adg        : ADC gain
    Return :
         0 : success
        -2 : failed to read list data
        -3 : failed to read measurement status
   ######################################################################### */
static int run_measure(FT_HANDLE usb_handle, USBMCA_HISTOGRAM *histdt, int *adg)
{
    USBMCA_STATUS   status;
    FT_STATUS ft_result;
    int     i, j;
    int     result ;
    int     ret0th;
    unsigned qword  meas_msec, meas_count;

    result = 0 ;
    
    /* save measurment info */
    memset(histdt, 0, sizeof(histdt[0]));
    histdt->meas_time = MEASURING_TIME;

    for (i = 0; i < (int)(_countof(histdt->adg)); i++) {
        histdt->adg[i] = usbmca_get_adg_maxch(adg[i]);
    }
    /************************/

    ft_result = usbmca_stop_measure(usb_handle);    /* measuring stop */

    meas_count = usbmca_msec2count(MEASURING_TIME); /* convert digit time */
    ft_result = usbmca_set_measuretime(usb_handle, meas_count);  /* setting measure time */
    ft_result = usbmca_set_measmode(usb_handle, USBMCA_MODE_HIST);       /* histogram mode */
    ft_result = usbmca_set_timemode(usb_handle, USBMCA_MODE_REALTIME);   /* real time */

    ft_result = usbmca_clear_data(usb_handle);      /* measurement data clear */
    ft_result = usbmca_start_measure(usb_handle);   /* measuring start */

    while (1) {
        /* read status */
        memset(&status, 0, sizeof(status));
        ret0th = usbmca_read_status(usb_handle, &status);

        if (ret0th < 0) {
            /* device read error(status)... */
            result = -1;
            break;
        }

        /* read histogram data */
        for (i = 0; i < (int)(_countof(histdt->data)); i++) {
            ret0th = usbmca_read_histogram(usb_handle, (i + 1), histdt->adg[i], histdt->data[i]);

            if (ret0th < 0) {
                /* device read error(histogram data)... */
                result = -2;
                break;
            }

            /* total count */
            histdt->totoal[i] = 0llu;

            for (j = 0; j < (int)(_countof(histdt->data[i])); j++) {
                histdt->totoal[i] += (unsigned qword)(histdt->data[i][j]);
            }
        }

        meas_msec = usbmca_count2msec(status.real_time);    /* convert msec */
        printf("elapsed time: %6llu msec\n", meas_msec);
        printf("    CH1: throuput rate = %6d, count = %6d.\n"
               "    CH2: throuput rate = %6d, count = %6d.\n"
               "    CH3: throuput rate = %6d, count = %6d.\n"
               "    CH4: throuput rate = %6d, count = %6d.\n\n",
            status.ch[0].rate.throughput, status.ch[0].count.throughput,
            status.ch[1].rate.throughput, status.ch[1].count.throughput,
            status.ch[2].rate.throughput, status.ch[2].count.throughput,
            status.ch[3].rate.throughput, status.ch[3].count.throughput);

        /* check measurement stop */
        if (meas_msec >= MEASURING_TIME) {
            file_save_histogram(SAVE_FILE, histdt);
            break;
        }

        msleep(250);    /* waiting */
    }

    usbmca_stop_measure(usb_handle) ;       /* stop */

    return  result ;
    
}


